import argparse
import gzip
import os
import uuid
from functools import partial
from multiprocessing import Pool

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from gtax.utils import chunks, sequences_to_list


def binning(transcript, bin, step):
    seq_len = len(transcript.seq)
    for i in range(1, seq_len - bin, step):
        yield SeqRecord(Seq(transcript.seq[i: i + bin]), id='{}_{}'.format(transcript.id, i),
                        description='')


def sequence_binning_to_tmpfile(transcripts, bin, step):
    pre = uuid.uuid4()
    with gzip.open('temp_{}.fa.gz'.format(pre), 'wt') as fout:
        sequences = {}
        for r in transcripts:
            for s in binning(r, bin, step):
                if not sequences.setdefault(str(s.seq), False):
                    sequences[str(s.seq)] = True
                    fout.write(s.format("fasta"))
    return pre


def sequence_binning(reference_file, output, seq_len, step, threads):
    print("Creating sequences")
    with Pool(processes=threads) as p:
        results = p.map(partial(sequence_binning_to_tmpfile,
                                bin=seq_len,
                                step=step),
                        [d for d in list(chunks(sequences_to_list(reference_file), 1000))])
    print('Printing final files')
    print('Creating file {}'.format(output))
    with gzip.open('{}'.format(output), 'wt') as fout2:
        sequences = {}
        for pre in results:
            with gzip.open('temp_{}.fa.gz'.format(pre), 'rt') as fin:
                for r in SeqIO.parse(fin, "fasta"):
                    if not sequences.setdefault(str(r.seq), False):
                        sequences[str(r.seq)] = True
                        fout2.write(r.format('fasta'))
            os.remove('temp_{}.fa.gz'.format(pre))


def sequence_binning_main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--reference', help='Reference Transcriptome gzip file',
                        required=True)
    parser.add_argument('--output', help='Output file prefix', required=True)
    parser.add_argument('--threads', help='Number of threads',
                        type=int, required=True)
    parser.add_argument('--seq_len', help='Length of the short sequence',
                        default=100, type=int, required=True)
    parser.add_argument('--step',
                        help='Sequence step to move the bin',
                        type=int, required=True)
    args = parser.parse_args()

    sequence_binning(args.reference,
                     args.output,
                     args.seq_len,
                     args.step,
                     args.threads)



import unittest

from gtax.taxonomy import Taxonomy


class TestTaxonomyMethods(unittest.TestCase):

    def test_taxonomy(self):
        self.taxonomy = Taxonomy()
        n = self.taxonomy.tax.nodes(data=True)['2']
        self.assertEqual(n['name'], 'Bacteria')
        # self.assertEqual('Bacteria', 'Bacteria')


if __name__ == '__main__':
    unittest.main()



import json
import os

from Bio import SeqIO


def use_sequence(accession):
    if accession.startswith('NW'):
        return False
    if accession.startswith('NZ'):
        if accession.startswith('NZ_CM') or accession.startswith('NZ_CP'):
            return True
        return False
    return True


def create_taxonomy_group_fasta(db):
    ids = {}

    if not os.path.exists('{}/ncbi_dataset/data/assembly_data_report.jsonl'.format(db)):
        raise Exception('File {}/ncbi_dataset/data/assembly_data_report.jsonl '
                        'don\'t exists'.format(db))

    print(db)
    fstream = open('{}.fsa'.format(db), 'w')
    fstream_tax = open('{}_taxid'.format(db), 'w')
    fstream_idx = open('{}.idx'.format(db), 'w')

    assemblies = {}
    with open('{}/ncbi_dataset/data/assembly_data_report.jsonl'.format(db)) as fjson:
        for line in fjson.readlines():
            d = json.loads(line)
            assemblies[d['assemblyInfo']['assemblyAccession']] = d['taxId']
    for s in assemblies:
        files = [f for ds, dr, files in os.walk('{}/ncbi_dataset/data/{}'.format(db, s))
                 for f in files if
                 f.endswith('.fna')]
        for f in files:
            for r in SeqIO.parse('{}/ncbi_dataset/data/{}/{}'.format(db, s, f), 'fasta'):
                seq_len = len(r.seq)
                if seq_len >= 1000 and use_sequence(r.id):
                    if not ids.setdefault(r.id, False):
                        ids[r.id] = True
                        fstream_tax.write('{}\t{}\n'.format(r.id, assemblies[s]))
                        fstream_idx.write('{}\t{}\t{}\t{}\n'.format(
                            r.id, fstream.tell(), assemblies[s], seq_len))
                        fstream.write(r.format('fasta'))
    fstream.close()
    fstream_tax.close()
    fstream_idx.close()
    print(db + ' done')


def create_taxonomy_group_eukaryotes_fasta(db, taxonomy):
    ids = {}

    if not os.path.exists('eukaryotes/ncbi_dataset/data/assembly_data_report.jsonl'):
        raise Exception('File eukaryotes/ncbi_dataset/data/assembly_data_report.jsonl'
                        ' don\'t exists')

    print(db)

    fstream = open('{}.fsa'.format(db), 'w')
    fstream_tax = open('{}_taxid'.format(db), 'w')
    fstream_idx = open('{}.idx'.format(db), 'w')

    assemblies = {}
    with open('eukaryotes/ncbi_dataset/data/assembly_data_report.jsonl') as fjson:
        for line in fjson.readlines():
            d = json.loads(line)
            if int(d['taxId']) in taxonomy.taxonomy_groups[db]['nodes']:
                assemblies[d['assemblyInfo']['assemblyAccession']] = d['taxId']
    for s in assemblies:
        files = [f for ds, dr, files in os.walk('eukaryotes/ncbi_dataset/data/{}'.format(s))
                 for f in files if
                 f.endswith('.fna')]
        for f in files:
            for r in SeqIO.parse('eukaryotes/ncbi_dataset/data/{}/{}'.format(s, f), 'fasta'):
                seq_len = len(r.seq)
                if seq_len >= 10000 and use_sequence(r.id):
                    if not ids.setdefault(r.id, False):
                        ids[r.id] = True
                        fstream_tax.write('{}\t{}\n'.format(r.id, assemblies[s]))
                        fstream_idx.write('{}\t{}\t{}\t{}\n'.format(
                            r.id, fstream.tell(), assemblies[s], seq_len))
                        fstream.write(r.format('fasta'))

    fstream.close()
    fstream_tax.close()
    fstream_idx.close()
    print(db + ' done')


def gtax_parallel(db, taxonomy):
    if db in ['archaea', 'bacteria', 'viruses']:
        create_taxonomy_group_fasta(db)
    else:
        create_taxonomy_group_eukaryotes_fasta(db, taxonomy)



from pkg_resources import DistributionNotFound

try:
    _distribution = __import__('pkg_resources').get_distribution("gtax")
except DistributionNotFound:  # Likely, running from working dir without installed dist
    __version__ = 'SNAPSHOT'
else:
    __version__ = _distribution.version if _distribution else 'SNAPSHOT'



import gzip

from Bio import SeqIO


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def yield_sequences(reference_file):
    with gzip.open(reference_file, 'rt') as trans_stream:
        for r in SeqIO.parse(trans_stream, "fasta"):
            yield r


def sequences_to_list(reference_file):
    sequences = []
    with gzip.open(reference_file, 'rt') as trans_stream:
        for r in SeqIO.parse(trans_stream, "fasta"):
            sequences.append(r)
    return sequences



import argparse
import gzip
import os
import random
import time
import uuid
from functools import partial
from multiprocessing import Pool

import pandas
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from gtax.taxonomy import Taxonomy
from gtax.utils import chunks


def load_reference_sequences(reference_file, total_sequences, short_seq_len):
    sequences = []
    with gzip.open(reference_file, 'rt') as trans_stream:
        for r in SeqIO.parse(trans_stream, "fasta"):
            sequences.append(r)
    print('{} transcripts loaded'.format(len(sequences)))
    return sequences


def load_reference_transcripts(reference_file, total_sequences, short_seq_len):
    transcripts = []
    sum_length = 0
    with gzip.open(reference_file, 'rt') as trans_stream:
        for r in SeqIO.parse(trans_stream, "fasta"):
            transcripts.append(r)
            sum_length += len(r.seq)
    print('{} transcripts loaded'.format(len(transcripts)))
    number_seq_factor = total_sequences * short_seq_len / sum_length
    print('Using factor {:.2f} to generate {} number of short sequences'.
          format(number_seq_factor, total_sequences))
    return transcripts, number_seq_factor


def fasta_paired_end_generator(records, number_seq_factor, short_seq_len, min_seq_size):
    pre = uuid.uuid4()
    print("Opening tmp files: temp_{0}_1.fa.gz and temp_{0}_1.fa.gz".format(pre))
    with gzip.open('temp_{}_1.fa.gz'.format(pre), 'wt') as fout_1, \
            gzip.open('temp_{}_2.fa.gz'.format(pre), 'wt') as fout_2:
        for r in records:
            length = len(r.seq)
            no_reads = int(number_seq_factor * length / short_seq_len)
            i = 0
            iterations = no_reads * 100
            while i < no_reads:
                iterations -= 1
                insert_size = random.randrange(200, 350)
                pos = random.randrange(length)
                f_start = pos - short_seq_len
                if f_start < 0:
                    f_start = 0
                f_read = r.seq[f_start:pos]
                r_start = pos + insert_size
                r_read = r.seq[r_start:r_start + short_seq_len].reverse_complement()
                if 'N' not in f_read and 'N' not in r_read and \
                        len(f_read) > min_seq_size and \
                        len(r_read) > min_seq_size:
                    f_rec = SeqRecord(Seq(f_read),
                                      id='{}_{}/1'.format(r.id, f_start), description='')
                    fout_1.write(f_rec.format("fasta"))

                    r_rec = SeqRecord(Seq(r_read),
                                      id='{}_{}/2'.format(r.id, f_start), description='')
                    fout_2.write(r_rec.format("fasta"))
                    i += 1
                if iterations <= 0:
                    break
    return pre


def create_random_paired_end_files(reference_file, output_prefix,
                                   number_sequences, short_seq_len, min_seq_size,
                                   number_files, threads, split_file_by):
    transcripts, number_seq_factor = load_reference_transcripts(reference_file,
                                                                number_sequences, short_seq_len)

    for i in range(1, number_files + 1):
        with Pool(processes=threads) as p:
            results = p.map(partial(fasta_paired_end_generator,
                                    number_seq_factor=number_seq_factor,
                                    short_seq_len=short_seq_len,
                                    min_seq_size=min_seq_size),
                            [d for d in list(chunks(transcripts, 1000))])

        print('Creating file {}_{}'.format(output_prefix, i))
        with gzip.open('{}_{}_1.fa.gz'.format(output_prefix, i), 'wt') as fout_1, \
                gzip.open('{}_{}_2.fa.gz'.format(output_prefix, i), 'wt') as fout_2, \
                gzip.open('{}_{}_cont_1.fa.gz'.format(output_prefix, i), 'wt') as fout2_1, \
                gzip.open('{}_{}_cont_2.fa.gz'.format(output_prefix, i), 'wt') as fout2_2:
            for pre in results:
                with gzip.open('temp_{}_1.fa.gz'.format(pre), 'rt') as fin:
                    for r in SeqIO.parse(fin, "fasta"):
                        fout_1.write(r.format('fasta'))
                        fout2_1.write(r.format('fasta'))
                os.remove('temp_{}_1.fa.gz'.format(pre))
                with gzip.open('temp_{}_2.fa.gz'.format(pre), 'rt') as fin:
                    for r in SeqIO.parse(fin, "fasta"):
                        fout_2.write(r.format('fasta'))
                        fout2_2.write(r.format('fasta'))
                os.remove('temp_{}_2.fa.gz'.format(pre))


def contamination_fasta_single_end_generator(tax_group, gtax_fasta_dir, df,
                                             total_read_seq, short_seq_len, min_seq_size):
    pre = uuid.uuid4()
    fasta_file = os.path.join(gtax_fasta_dir, '{}.fsa'.format(tax_group))
    output_file = 'temp_{}_{}.fa.gz'.format(tax_group, pre)
    with open(fasta_file) as fin, \
            gzip.open(output_file, 'wt') as fout:
        for idx, row in df.iterrows():
            fin.seek(row[1], 0)
            for r in SeqIO.parse(fin, "fasta"):
                length = len(r.seq)
                i = 0
                iterations = total_read_seq * 1000
                while i < total_read_seq:
                    iterations -= 1
                    pos = random.randrange(length)
                    f_start = pos - short_seq_len
                    if f_start < 0:
                        f_start = 0
                    f_read = r.seq[f_start:pos]
                    if 'N' not in f_read and len(f_read) > min_seq_size:
                        f_rec = SeqRecord(Seq(f_read),
                                          id='{}_{}'.format(r.id, f_start), description='')
                        fout.write(f_rec.format("fasta"))
                        i += 1
                    if iterations <= 0:
                        break
                break
    return output_file


def process_by_taxonomy_group(tax_group, gtax_fasta_dir, total_reads, short_seq_len, min_seq_size):
    idx_file = os.path.join(gtax_fasta_dir, '{}.idx'.format(tax_group))
    df = pandas.read_csv(idx_file, sep='\t', header=None)
    total_seq = len(df)
    total_read_seq = int(total_reads / total_seq)
    if total_read_seq < 1000:
        total_read_seq = 1000
        total_seq = int(total_reads / total_read_seq)
        df = df.sample(total_seq, random_state=random.randint(1, int(time.time()))).sort_values(by=1)
    return contamination_fasta_single_end_generator(tax_group, gtax_fasta_dir,
                                                    df,
                                                    total_read_seq,
                                                    short_seq_len, min_seq_size)


def fasta_single_end_generator(records, number_seq_factor, short_seq_len, min_seq_size):
    pre = uuid.uuid4()
    with gzip.open('temp_{}.fa.gz'.format(pre), 'wt') as fout:
        for r in records:
            length = len(r.seq)
            no_reads = int(number_seq_factor * length / short_seq_len)
            i = 0
            iterations = no_reads * 1000
            while i < no_reads:
                iterations -= 1
                pos = random.randrange(length)
                f_start = pos - short_seq_len
                if f_start < 0:
                    f_start = 0
                f_read = r.seq[f_start:pos]
                if 'N' not in f_read and len(f_read) > min_seq_size:
                    f_rec = SeqRecord(Seq(f_read), id='{}_{}'.format(r.id, f_start), description='')
                    fout.write(f_rec.format("fasta"))
                    i += 1
                if iterations <= 0:
                    break
    return pre


def create_random_single_end_files(reference_file, output_prefix,
                                   number_sequences, short_seq_len, min_seq_size,
                                   number_files, threads, split_file_by,
                                   taxonomy, foreign_contamination_percent,
                                   gtax_fasta_dir, taxonomy_group):
    print("Creating single-end short sequences")
    ref_number_sequences = number_sequences - \
                           (number_sequences * foreign_contamination_percent *
                            (len(taxonomy.taxonomy_groups) - 1))
    transcripts, number_seq_factor = load_reference_transcripts(reference_file,
                                                                ref_number_sequences, short_seq_len)

    for i in range(1, number_files + 1):
        print('Creating reference sequences')
        with Pool(processes=threads) as p:
            results = p.map(partial(fasta_single_end_generator,
                                    number_seq_factor=number_seq_factor,
                                    short_seq_len=short_seq_len,
                                    min_seq_size=min_seq_size),
                            [d for d in list(chunks(transcripts, 1000))])
        print('Creating foreign sequences')
        with Pool(processes=threads) as p:
            results_cont = p.map(partial(process_by_taxonomy_group,
                                         gtax_fasta_dir=gtax_fasta_dir,
                                         total_reads=number_sequences * foreign_contamination_percent,
                                         short_seq_len=short_seq_len,
                                         min_seq_size=min_seq_size),
                                 [d for d in taxonomy.taxonomy_groups if d != taxonomy_group])
        print('Printing final files')
        if split_file_by == 0:
            print('Creating file {}_{}'.format(output_prefix, i))
            with gzip.open('{}_{}.fa.gz'.format(output_prefix, i), 'wt') as fout2:
                for pre in results:
                    with gzip.open('temp_{}.fa.gz'.format(pre), 'rt') as fin:
                        for r in SeqIO.parse(fin, "fasta"):
                            fout2.write(r.format('fasta'))
                    os.remove('temp_{}.fa.gz'.format(pre))
                for pre in results_cont:
                    with gzip.open(pre, 'rt') as fin:
                        for r in SeqIO.parse(fin, "fasta"):
                            fout2.write(r.format('fasta'))
                    os.remove(pre)
        else:
            count = 0
            suffix = 1
            fout2 = gzip.open('{}_{}_{}.fa.gz'.format(output_prefix, i, suffix), 'wt')
            for pre in results:
                with gzip.open('temp_{}.fa.gz'.format(pre), 'rt') as fin:
                    for r in SeqIO.parse(fin, "fasta"):
                        if count == split_file_by:
                            fout2.close()
                            count = 0
                            suffix += 1
                            fout2 = gzip.open('{}_{}_{}_cont.fa.gz'.format(output_prefix, i, suffix), 'wt')
                        count += 1
                        fout2.write(r.format('fasta'))
                os.remove('temp_{}.fa.gz'.format(pre))
            for pre in results_cont:
                with gzip.open(pre, 'rt') as fin:
                    for r in SeqIO.parse(fin, "fasta"):
                        if count == split_file_by:
                            fout2.close()
                            count = 0
                            suffix += 1
                            fout2 = gzip.open('{}_{}_{}_cont.fa.gz'.format(output_prefix, i, suffix), 'wt')
                        count += 1
                        fout2.write(r.format('fasta'))
                os.remove(pre)
            fout2.close()


def create_random_short_sequences():
    parser = argparse.ArgumentParser()

    parser.add_argument('--reference', help='Reference Transcriptome gzip file',
                        required=True)
    parser.add_argument('--taxid', help='TaxID for target organism',
                        type=int,
                        required=True)
    parser.add_argument('--gtax_fasta_dir', help='Gtax database directory contaning FASTA and IDX files',
                        required=True)
    parser.add_argument('--output_prefix', help='Output file prefix', required=True)
    parser.add_argument('--number_files', help='Number of random file to create',
                        default=4, type=int, required=False)
    parser.add_argument('--threads', help='Number of threads',
                        type=int, required=True)
    parser.add_argument('--foreign_contamination_percent',
                        help='Foreign Contamination percent by taxonomy group',
                        type=float, required=True)
    parser.add_argument('--short_seq_len', help='Length of the short read',
                        default=100, type=int, required=False)
    parser.add_argument('--number_sequences',
                        help='The number of short reads sequences to generate ',
                        type=int, required=True)
    parser.add_argument('--split_file_by',
                        help='Generate output files with this number of sequences ',
                        default=0,
                        type=int, required=False)
    parser.add_argument('--paired_end',
                        help='Generate paired-end mate sequence',
                        action="store_true", default=False, required=False)
    args = parser.parse_args()
    min_seq_size = args.short_seq_len * 75 / 100

    taxonomy = Taxonomy(tax_pickle_file=os.path.join(args.gtax_fasta_dir,
                                                     'taxonomy.pickle'),
                        group_pickle_file=os.path.join(args.gtax_fasta_dir,
                                                       'taxonomy_groups.pickle'))

    taxonomy_group = taxonomy.get_taxonomy_group_from_taxid(args.taxid)
    if taxonomy_group:
        print('TaxID {} is in {} taxonomy group'.format(args.taxid, taxonomy_group))

        if args.paired_end:
            create_random_paired_end_files(args.reference,
                                           args.output_prefix,
                                           args.number_sequences,
                                           args.short_seq_len,
                                           min_seq_size,
                                           args.number_files,
                                           args.threads,
                                           args.split_file_by)
        else:
            create_random_single_end_files(args.reference,
                                           args.output_prefix,
                                           args.number_sequences,
                                           args.short_seq_len,
                                           min_seq_size,
                                           args.number_files,
                                           args.threads,
                                           args.split_file_by,
                                           taxonomy,
                                           args.foreign_contamination_percent,
                                           args.gtax_fasta_dir,
                                           taxonomy_group)
    else:
        print('Reference organism is not in the database')



import os
import os.path

from setuptools import find_packages
from setuptools import setup


def readme():
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
        return f.read()

# Set __version__
exec(open('src/gtax/__init__.py').read())

setup(
    name='gtax',
    packages=find_packages(where='src'),
    package_dir={
        '': 'src',
    },
    data_files=[('', ['README.md'])],
    use_scm_version=True,
    setup_requires=['wheel', 'setuptools_scm'],
    description='Gtax generate a taxonomy sequence database by genomes ',
    long_description=readme(),
    long_description_content_type='text/markdown',
    license='Public Domain',
    author='Vera Alvarez, Roberto',
    author_email='veraalva' '@' 'ncbi.nlm.nih.gov',
    maintainer='Vera Alvarez, Roberto',
    maintainer_email='veraalva' '@' 'ncbi.nlm.nih.gov',
    url='https://github.com/ncbi/gtax',
    install_requires=['biopython',
                      'networkx',
                      'numpy',
                      'pandas',
                      'requests'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    keywords='Biocontainers',
    project_urls={
        'Documentation': 'https://gtax.readthedocs.io/',
        'Source': 'https://github.com/ncbi/gtax',
        'Tracker': 'https://github.com/ncbi/gtax/issues',
    },
    entry_points={
        'console_scripts': [
            'taxonomy_pickle = gtax.taxonomy_main:taxonomy_pickle',
            'gtax_database = gtax.gtax_main:gtax_database',
            'filter_metadata_zip = gtax.gtax_main:filter_metadata_zip',
            'create_random_short_sequences = gtax.sequence:create_random_short_sequences',
            'sequence_binning = gtax.sequence_binning:sequence_binning_main'
        ],
    }
)



import unittest

from gtax.gtax_main import check_preexisting_data


class TestGtaxMethods(unittest.TestCase):

    def test_preexisting_data(self):
        self.assertRaises(Exception, check_preexisting_data)


if __name__ == '__main__':
    unittest.main()



import json
import os
from functools import partial
from multiprocessing import Pool
from zipfile import ZipFile

from gtax.gtax_database import gtax_parallel
from gtax.taxonomy import Taxonomy


def check_preexisting_data():
    dbs = ['archaea', 'bacteria', 'eukaryotes', 'viruses']
    for db in dbs:
        if not os.path.exists('{}/ncbi_dataset/data/assembly_data_report.jsonl'.format(db)):
            raise Exception('Folder {} does not exists. '
                            'Use datasets for downloading the data'.format(db))


def gtax_database():
    check_preexisting_data()
    taxonomy = Taxonomy()
    dbs = taxonomy.taxonomy_groups.keys()
    print('Processing {} databases'.format(dbs))
    with Pool(len(dbs)) as p:
        p.map(partial(gtax_parallel, taxonomy=taxonomy), dbs)
    taxonomy.create_taxonomy_groups()
    taxonomy.create_pickle('taxonomy.pickle',
                           'taxonomy_groups.pickle')


def filter_metadata_zip():
    superkingdoms = ['archaea', 'bacteria', 'viruses', 'eukaryotes']

    for db in superkingdoms:
        if os.path.exists('{}_meta.zip'.format(db)):
            if not os.path.exists('{}/ncbi_dataset/data'.format(db)):
                os.makedirs('{}/ncbi_dataset/data'.format(db))
            with ZipFile('{}_meta.zip'.format(db), 'r') as zip:
                assemblies = []
                assemblies_tmp = {}
                with zip.open('ncbi_dataset/data/assembly_data_report.jsonl') as fjson, open(
                        '{}/ncbi_dataset/data/assembly_data_report.jsonl'.format(db), 'w') as fjson_out:
                    for line in fjson.readlines():
                        d = json.loads(line.decode("utf-8"))
                        v = assemblies_tmp.setdefault(d['taxId'], [])
                        v.append(d)
                    for s in assemblies_tmp.keys():
                        rep_genome = []
                        for e in assemblies_tmp[s]:
                            if 'refseqCategory' in e['assemblyInfo']:
                                rep_genome.append(e)
                        if len(rep_genome) == 1:
                            assemblies.append(rep_genome[0]['assemblyInfo']['assemblyAccession'])
                            fjson_out.write('{}\n'.format(json.dumps(rep_genome[0])))
                        else:
                            assemblies.append(assemblies_tmp[s][0]['assemblyInfo']['assemblyAccession'])
                            fjson_out.write('{}\n'.format(json.dumps(assemblies_tmp[s][0])))

                print('There are {} assemblies included'.format(len(assemblies)))
                with zip.open('ncbi_dataset/data/dataset_catalog.json') as fjson, open(
                        '{}/ncbi_dataset/data/dataset_catalog.json'.format(db), 'w') as fjson_out:
                    d = json.loads(fjson.read().decode("utf-8"))
                    catalog = []
                    for c in d['assemblies']:
                        if 'accession' in c:
                            if c['accession'] in assemblies:
                                catalog.append(c)
                        else:
                            catalog.append(c)
                    d['assemblies'] = catalog
                    fjson_out.write(json.dumps(d, indent=2))
                with zip.open('ncbi_dataset/fetch.txt') as fin, open('{}/ncbi_dataset/fetch.txt'.format(db), 'w') as fout:
                    for line in fin.readlines():
                        line = line.decode("utf-8")
                        f = os.path.dirname(line.split('\t')[2].replace('data/', ''))
                        if f in assemblies:
                            fout.write(line)



import os
import pickle
import shutil
import tarfile
import tempfile

import networkx as nx
import pandas
import requests


def parse_nodes_file(node_file, taxid):
    with open(node_file, 'r') as fin:
        for line in fin:
            f = line.strip().split('\t|\t')
            node = {}
            node['id'] = f[0]
            node['name'] = taxid[f[0]]['scientific name']
            edge = ()
            if f[1] != node['id']:
                edge = (f[1], node['id'])
            yield (f[0], node), edge


def parse_tax_name_file(name_file):
    tax_id = {}
    with open(name_file, 'r') as fin:
        for line in fin:
            line = line.strip()
            f = line.split('\t|\t')
            if f[0] not in tax_id:
                tax_id[f[0]] = {}
            tax_id[f[0]][f[3].replace('\t|', '').strip()] = f[1]
    return tax_id


class Taxonomy:

    def __init__(self, *args, **kwargs):
        name_file = kwargs.get('name_file', None)
        node_file = kwargs.get('node_file', None)
        tax_pickle_file = kwargs.get('tax_pickle_file', None)
        group_pickle_file = kwargs.get('group_pickle_file', None)
        self.taxonomy_groups = kwargs.get('taxonomy_groups', {
            'bacteria': {'taxid': '2', 'nodes': set(), 'sequences': set(), 'size': 0},
            'archaea': {'taxid': '2157', 'nodes': set(), 'sequences': set(), 'size': 0},
            'liliopsida': {'taxid': '4447', 'nodes': set(), 'sequences': set(), 'size': 0},
            'eudicotyledons': {'taxid': '71240', 'nodes': set(), 'sequences': set(), 'size': 0},
            'viridiplantae': {'taxid': '33090', 'nodes': set(), 'sequences': set(), 'size': 0},
            'fungi': {'taxid': '4751', 'nodes': set(), 'sequences': set(), 'size': 0},
            'arthropoda': {'taxid': '6656', 'nodes': set(), 'sequences': set(), 'size': 0},
            'neoteleostei': {'taxid': '123365', 'nodes': set(), 'sequences': set(), 'size': 0},
            'actinopterygii': {'taxid': '7898', 'nodes': set(), 'sequences': set(), 'size': 0},
            'glires': {'taxid': '314147', 'nodes': set(), 'sequences': set(), 'size': 0},
            'primates': {'taxid': '9443', 'nodes': set(), 'sequences': set(), 'size': 0},
            'carnivora': {'taxid': '33554', 'nodes': set(), 'sequences': set(), 'size': 0},
            'artiodactyla': {'taxid': '91561', 'nodes': set(), 'sequences': set(), 'size': 0},
            'amphibia': {'taxid': '8292', 'nodes': set(), 'sequences': set(), 'size': 0},
            'sauropsida': {'taxid': '8457', 'nodes': set(), 'sequences': set(), 'size': 0},
            'sarcopterygii': {'taxid': '8287', 'nodes': set(), 'sequences': set(), 'size': 0},
            'chordata': {'taxid': '7711', 'nodes': set(), 'sequences': set(), 'size': 0},
            'eukaryota': {'taxid': '2759', 'nodes': set(), 'sequences': set(), 'size': 0},
            'viruses': {'taxid': '10239', 'nodes': set(), 'sequences': set(), 'size': 0},
        })
        self.nodes = None
        self.tax = nx.DiGraph()
        if name_file and node_file:
            self.create_taxonomy_graph(name_file, node_file)
            self.create_taxonomy_groups()
        elif tax_pickle_file:
            self.tax = pickle.load(open(tax_pickle_file, "rb"))
            print('{} taxonomies loaded'.format(len(self.tax.nodes())))
            self.nodes = self.tax.nodes(data=True)
            self.taxonomy_groups = pickle.load(open(group_pickle_file, "rb"))
            for k in self.taxonomy_groups:
                print('{} Node: {} Sequences: {}'.format(
                    k, len(self.taxonomy_groups[k]['nodes']),
                    len(self.taxonomy_groups[k]['sequences'])
                ))
        else:
            self.create_taxonomy_from_data()
            self.create_taxonomy_groups()

    def create_taxonomy_from_data(self):
        print('Downloading NCBI Taxonomy database')
        dirpath = tempfile.mkdtemp()
        url = 'https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz'
        response = requests.get(url, stream=True)
        file = tarfile.open(fileobj=response.raw, mode="r|gz")
        file.extractall(path=dirpath)
        self.create_taxonomy_graph(os.path.join(dirpath, 'names.dmp'),
                                   os.path.join(dirpath, 'nodes.dmp'))

        shutil.rmtree(dirpath)

    def successors(self, g):
        """
        Extract ancestors nodes from an starting node
        :param g: starting node name
        :param O: Graph
        :return: a set with node names
        """
        result = {g}
        for o in self.tax.successors(g):
            result.update(self.successors(o))
        return result

    def find_node(self, id):
        nodes = [y for x, y in self.nodes if y['id'] == id]
        if nodes:
            a = ""
            for i in nx.shortest_path(self.tax, source="1", target=id)[2:]:
                ns = [y for x, y in self.nodes if y['id'] == i]
                if ns:
                    if a:
                        a += "; "
                    a += ns[0]['name']
            return nodes[0], a
        return None, None

    def find_node_by_name(self, name):
        for n in self.nodes:
            if n[1]['name'].casefold() == name.casefold():
                return n
        return None

    def get_node_lineage(self, node):
        a = ""
        for i in nx.shortest_path(self.tax, source="1", target=node[0])[2:]:
            ns = [y for x, y in self.nodes if y['id'] == i]
            if ns:
                if a:
                    a += "; "
                a += ns[0]['name']
        return a

    def get_lineage_by_name(self, name):
        n = self.find_node_by_name(name)
        return self.get_node_lineage(n)

    def create_taxonomy_graph(self, name_file, node_file):
        tax_id = parse_tax_name_file(name_file)
        print('Taxonomies: {}'.format(len(tax_id)))
        entries = parse_nodes_file(node_file, tax_id)
        nodes, edges = zip(*entries)
        print('{} nodes created'.format(len(nodes)))
        self.tax.add_nodes_from(nodes)
        for e in edges:
            if e:
                self.tax.add_edge(*e)
        self.nodes = self.tax.nodes(data=True)

    def get_successors(self, taxid):
        return set([int(i) for i in self.successors(taxid)])

    def get_taxonomy_group_nodes(self, taxid, to_exclude):
        self.taxonomy_groups[taxid]['nodes'] = \
            self.get_successors(self.taxonomy_groups[taxid]['taxid'])
        self.taxonomy_groups[taxid]['nodes'] = \
            list(set(self.taxonomy_groups[taxid]['nodes']) - to_exclude)
        print('{} with {} taxa'.format(taxid, len(self.taxonomy_groups[taxid]['nodes'])))

    def add_sequences_size_from_gtax_idx(self, taxonomy_group):
        if os.path.exists('{}.idx'.format(taxonomy_group)):
            df = pandas.read_csv('{}.idx'.format(taxonomy_group), sep='\t', header=None)
            print('{} sequences loaded from the index'.format(len(df)))
            seq = {}
            for g in df.groupby(2):
                seq[str(g[0])] = {'sequences': set(g[1][0].unique()), 'size': g[1][3].sum()}
            nx.set_node_attributes(self.tax, seq)

    def add_sequences_size_from_gtax_idx_all(self):
        for k in self.taxonomy_groups:
            self.add_sequences_size_from_gtax_idx(k)

    def create_taxonomy_groups(self):
        inserted = set()
        for k in self.taxonomy_groups:
            self.taxonomy_groups[k]['nodes'] = \
                self.get_successors(self.taxonomy_groups[k]['taxid']).difference(
                    inserted)
            inserted.update(self.taxonomy_groups[k]['nodes'])
            self.add_sequences_size_from_gtax_idx(k)
            for node_id in self.taxonomy_groups[k]['nodes']:
                node_id = str(node_id)
                if 'size' in self.nodes[node_id]:
                    self.taxonomy_groups[k]['sequences'].update(self.nodes[node_id]['sequences'])
                    self.taxonomy_groups[k]['size'] += self.nodes[node_id]['size']

    def create_pickle(self, tax_pickle_file, group_pickle_file):
        print('Printing tax graph pickle file: {}'.format(tax_pickle_file))
        pickle.dump(self.tax, open(tax_pickle_file, "wb"))
        print('Printing tax group pickle file: {}'.format(group_pickle_file))
        pickle.dump(self.taxonomy_groups, open(group_pickle_file, "wb"))

    def print_size(self, node_name, deep=1, step=1, min_size=10.0, min_size_child=50.0):
        node = self.find_node_by_name(node_name)
        size = 0
        taxa = 0
        for node_id in self.successors(node[1]['id']):
            if 'sequences' in self.nodes[node_id]:
                taxa += 1
                size += self.nodes[node_id]['size']
        size = size / 1e+9
        if size >= min_size:
            print('{}{} {} => {:.2f} GB'.format('   ' * step, node_name, node[1]['id'], size))
        step += 1
        if step < deep and size >= min_size_child:
            for t in self.tax.successors(node[1]['id']):
                if t != node[1]['id']:
                    self.print_size(self.nodes[str(t)]['name'],
                                    deep, step, min_size, min_size_child)

    def get_taxonomy_group_from_taxid(self, taxid):
        for k in self.taxonomy_groups:
            if taxid in self.taxonomy_groups[k]['nodes']:
                return k
        return None